from pathlib import Path
from argparse import Namespace
from typing import List, Optional, Union

from jina.helper import colored, random_port
from jina.peapods import Pea, Pod, CompoundPod
from jina.peapods.peas.helper import update_runtime_cls
from jina import Flow, __docker_host__
from jina.logging.logger import JinaLogger

from .. import jinad_args, __partial_workspace__
from ..models import GATEWAY_RUNTIME_DICT
from ..models.enums import UpdateOperation
from ..models.ports import Ports, PortMappings
from ..models.partial import PartialFlowItem, PartialStoreItem


class PartialStore:
    """A store spawned inside partial-daemon container"""

    def __init__(self):
        self._logger = JinaLogger(self.__class__.__name__, **vars(jinad_args))
        self.item = PartialStoreItem()

    def add(self, *args, **kwargs) -> PartialStoreItem:
        """Add a new element to the store. This method needs to be overridden by the subclass


        .. #noqa: DAR101"""
        raise NotImplementedError

    def update(self, *args, **kwargs) -> PartialStoreItem:
        """Updates the element to the store. This method needs to be overridden by the subclass


        .. #noqa: DAR101"""
        raise NotImplementedError

    def delete(self) -> None:
        """Terminates the object in the store & stops the server"""
        try:
            if hasattr(self, 'object'):
                self.object.close()
            else:
                self._logger.warning(f'nothing to close. exiting')
        except Exception as e:
            self._logger.error(f'{e!r}')
            raise


class PartialPeaStore(PartialStore):
    """A Pea store spawned inside partial-daemon container"""

    peapod_cls = Pea

    def add(self, args: Namespace, **kwargs) -> PartialStoreItem:
        """Starts a Pea in `partial-daemon`

        :param args: namespace args for the pea/pod
        :param kwargs: keyword args
        :return: Item describing the Pea object
        """
        try:
            _id = args.identity
            # This is set so that ContainerRuntime sets host_ctrl to __docker_host__
            # and on linux machines, we can access dockerhost inside containers
            if args.runtime_cls == 'ContainerRuntime':
                args.docker_kwargs = {'extra_hosts': {__docker_host__: 'host-gateway'}}
            self.object: Union['Pea', 'Pod'] = self.peapod_cls(args).__enter__()
        except Exception as e:
            if hasattr(self, 'object'):
                self.object.__exit__(type(e), e, e.__traceback__)
            self._logger.error(f'{e!r}')
            raise
        else:
            self.item = PartialStoreItem(arguments=vars(args))
            self._logger.success(f'{colored(_id, "cyan")} is created')
            return self.item


class PartialPodStore(PartialPeaStore):
    """A Pod store spawned inside partial-daemon container"""

    peapod_cls = Pod


class PartialFlowStore(PartialStore):
    """A Flow store spawned inside partial-daemon container"""

    def add(
        self, args: Namespace, port_mapping: Optional[PortMappings] = None, **kwargs
    ) -> PartialStoreItem:
        """Starts a Flow in `partial-daemon`.

        :param args: namespace args for the flow
        :param port_mapping: ports to be set
        :param kwargs: keyword args
        :return: Item describing the Flow object
        """
        try:
            if not args.uses:
                raise ValueError('uses yaml file was not specified in flow definition')
            elif not Path(args.uses).is_file():
                raise ValueError(f'uses {args.uses} not found in workspace')

            with open(args.uses) as yaml_file:
                yaml_source = yaml_file.read()

            self.object: Flow = Flow.load_config(yaml_source).build()
            self.object.workspace_id = jinad_args.workspace_id
            self.object.workspace = __partial_workspace__
            self.object.env = {'HOME': __partial_workspace__}
            # TODO(Deepankar): pass envs from main daemon process to partial-daemon so that
            # Pods/Peas/Runtimes/Executors can inherit these env vars

            for pod in self.object._pod_nodes.values():
                runtime_cls = update_runtime_cls(pod.args, copy=True).runtime_cls
                if isinstance(pod, CompoundPod):
                    # In dependencies, we set `runs_in_docker` for the `gateway` and for `CompoundPod` we need
                    # `runs_in_docker` to be False. Since `Flow` args are sent to all Pods, `runs_in_docker` gets set
                    # for the `CompoundPod`, which blocks the requests. Below we unset that (hacky & ugly).
                    # We do it only for runtimes that starts on local (not container or remote)
                    if (
                        runtime_cls
                        in [
                            'ZEDRuntime',
                            'GRPCDataRuntime',
                            'ContainerRuntime',
                        ]
                        + list(GATEWAY_RUNTIME_DICT.values())
                    ):
                        pod.args.runs_in_docker = False
                        for shards_args in pod.shards_args:
                            shards_args.runs_in_docker = False
                        if port_mapping:
                            # Ports for Head & Tail Peas in a CompoundPod set here.
                            # This is specifically needed as `save_config` doesn't save `port_out` for a HeadPea
                            # and `port_in` for a TailPea, which might be useful if replicas and head/tail Peas
                            # are in different containers.
                            for pea_args in [pod.head_args, pod.tail_args]:
                                if pea_args.name in port_mapping.pea_names:
                                    for port_name in Ports.__fields__:
                                        self._set_pea_ports(
                                            pea_args, port_mapping, port_name
                                        )
                            # Update shard_args according to updated head & tail args
                            pod.assign_shards()
                else:
                    if port_mapping and (
                        hasattr(pod.args, 'replicas') and pod.args.replicas > 1
                    ):
                        for pea_args in [pod.peas_args['head'], pod.peas_args['tail']]:
                            if pea_args.name in port_mapping.pea_names:
                                for port_name in Ports.__fields__:
                                    self._set_pea_ports(
                                        pea_args, port_mapping, port_name
                                    )
                        pod.update_worker_pea_args()

                    # avoid setting runs_in_docker for Pods with parallel > 1 and using `ZEDRuntime`
                    # else, replica-peas would try connecting to head/tail-pea via __docker_host__
                    if runtime_cls in ['ZEDRuntime', 'GRPCDataRuntime'] and (
                        hasattr(pod.args, 'replicas') and pod.args.replicas > 1
                    ):
                        pod.args.runs_in_docker = False
                        pod.update_worker_pea_args()

            self.object = self.object.__enter__()
        except Exception as e:
            if hasattr(self, 'object'):
                self.object.__exit__(type(e), e, e.__traceback__)
            self._logger.error(f'{e!r}')
            raise
        else:
            self.item = PartialFlowItem(
                arguments={
                    'port_expose': self.object.port_expose,
                    'protocol': self.object.protocol.name,
                    **vars(self.object.args),
                },
                yaml_source=yaml_source,
            )
            self._logger.success(f'Flow is created')
            return self.item

    def _set_pea_ports(self, pea_args, port_mapping, port_name):
        if hasattr(pea_args, port_name):
            setattr(
                pea_args,
                port_name,
                getattr(
                    port_mapping[pea_args.name].ports,
                    port_name,
                    random_port(),
                ),
            )

    def update(
        self,
        kind: UpdateOperation,
        dump_path: str,
        pod_name: str,
        shards: int,
        **kwargs,
    ) -> PartialFlowItem:
        """Runs an update operation on the Flow.
        :param kind: type of update command to execute (dump/rolling_update)
        :param dump_path: the path to which to dump on disk
        :param pod_name: pod to target with the dump request
        :param shards: nr of shards to dump
        :param kwargs: keyword args
        :return: Item describing the Flow object
        """
        try:
            if kind == UpdateOperation.ROLLING_UPDATE:
                self.object.rolling_update(pod_name=pod_name, dump_path=dump_path)
            else:
                self._logger.error(f'unsupoorted kind: {kind}, no changes done')
                return self.item
        except Exception as e:
            self._logger.error(f'{e!r}')
            raise
        else:
            self.item.arguments = vars(self.object.args)
            return self.item



import argparse
import asyncio
from typing import Optional, Callable

import grpc
from google.protobuf import struct_pb2

from jina.logging.logger import JinaLogger
from jina.peapods.networking import create_connection_pool

from jina.proto import jina_pb2_grpc, jina_pb2
from jina.types.message import Message
from jina.types.message.common import ControlMessage
from jina.types.routing.table import RoutingTable


class Grpclet(jina_pb2_grpc.JinaDataRequestRPCServicer):
    """A `Grpclet` object can send/receive Messages via gRPC.

    :param args: the parsed arguments from the CLI
    :param message_callback: the callback to call on received messages
    :param logger: the logger to use
    """

    def __init__(
        self,
        args: argparse.Namespace,
        message_callback: Callable[['Message'], None],
        logger: Optional['JinaLogger'] = None,
    ):
        self.args = args
        self._logger = logger or JinaLogger(self.__class__.__name__)
        self.callback = message_callback

        self._connection_pool = create_connection_pool(args)

        self.msg_recv = 0
        self.msg_sent = 0
        self._pending_tasks = []
        self._static_routing_table = args.static_routing_table
        if args.static_routing_table:
            self._routing_table = RoutingTable(args.routing_table)
            self._next_targets = self._routing_table.get_next_target_addresses()
        else:
            self._routing_table = None
            self._next_targets = None

    async def send_message(self, msg: 'Message', **kwargs):
        """
        Sends a message via gRPC to the target indicated in the message's routing table
        :param msg: the protobuf message to send
        :param kwargs: Additional arguments.
        """

        if self._next_targets:
            for pod_address in self._next_targets:
                self._send_message(msg, pod_address)
        else:
            routing_table = RoutingTable(msg.envelope.routing_table)
            next_targets = routing_table.get_next_targets()
            for target, _ in next_targets:
                self._send_message(
                    self._add_envelope(msg, target),
                    target.active_target_pod.full_address,
                )

    def _send_message(self, msg, pod_address):
        try:
            self.msg_sent += 1

            self._pending_tasks.append(
                self._connection_pool.send_message(msg, pod_address)
            )

            self._update_pending_tasks()
        except grpc.RpcError as ex:
            self._logger.error('Sending data request via grpc failed', ex)
            raise ex

    @staticmethod
    def send_ctrl_msg(pod_address: str, command: str, timeout=1.0):
        """
        Sends a control message via gRPC to pod_address
        :param pod_address: the pod to send the command to
        :param command: the command to send (TERMINATE/ACTIVATE/...)
        :param timeout: optional timeout for the request in seconds
        :returns: Empty protobuf struct
        """
        stub = Grpclet._create_grpc_stub(pod_address, is_async=False)
        response = stub.Call(ControlMessage(command), timeout=timeout)
        return response

    @staticmethod
    def _create_grpc_stub(pod_address, is_async=True):
        if is_async:
            channel = grpc.aio.insecure_channel(
                pod_address,
                options=[
                    ('grpc.max_send_message_length', -1),
                    ('grpc.max_receive_message_length', -1),
                ],
            )
        else:
            channel = grpc.insecure_channel(
                pod_address,
                options=[
                    ('grpc.max_send_message_length', -1),
                    ('grpc.max_receive_message_length', -1),
                ],
            )

        stub = jina_pb2_grpc.JinaDataRequestRPCStub(channel)

        return stub

    def _add_envelope(self, msg, routing_table):
        if not self._static_routing_table:
            new_envelope = jina_pb2.EnvelopeProto()
            new_envelope.CopyFrom(msg.envelope)
            new_envelope.routing_table.CopyFrom(routing_table.proto)
            new_message = Message(request=msg.request, envelope=new_envelope)

            return new_message
        else:
            return msg

    async def close(self, grace_period=None, *args, **kwargs):
        """Stop the Grpc server
        :param grace_period: Time to wait for message processing to finish before killing the grpc server
        :param args: Extra positional arguments
        :param kwargs: Extra key-value arguments
        """
        self._update_pending_tasks()

        try:
            await asyncio.wait_for(asyncio.gather(*self._pending_tasks), timeout=1.0)
        except asyncio.TimeoutError:
            self._update_pending_tasks()
            self._logger.warning(
                f'Could not gracefully complete {len(self._pending_tasks)} pending tasks on close.'
            )

        self._connection_pool.close()
        self._logger.debug('Close grpc server')
        await self._grpc_server.stop(grace_period)

    async def start(self):
        """
        Starts this Grpclet by starting its gRPC server
        """
        self._grpc_server = grpc.aio.server(
            options=[
                ('grpc.max_send_message_length', -1),
                ('grpc.max_receive_message_length', -1),
            ]
        )

        jina_pb2_grpc.add_JinaDataRequestRPCServicer_to_server(self, self._grpc_server)
        bind_addr = f'{self.args.host}:{self.args.port_in}'
        self._grpc_server.add_insecure_port(bind_addr)
        self._logger.debug(f'Binding gRPC server for data requests to {bind_addr}')
        self._connection_pool.start()
        await self._grpc_server.start()
        await self._grpc_server.wait_for_termination()

    def _update_pending_tasks(self):
        self._pending_tasks = [
            task for task in self._pending_tasks if task and not task.done()
        ]

    async def Call(self, msg, *args):
        """Processes messages received by the GRPC server
        :param msg: The received message
        :param args: Extra positional arguments
        :return: Empty protobuf struct, necessary to return for protobuf Empty
        """
        if self.callback:
            self._pending_tasks.append(asyncio.create_task(self.callback(msg)))
        else:
            self._logger.debug(
                'Grpclet received data request, but no callback was registered'
            )

        self.msg_recv += 1
        self._update_pending_tasks()
        return struct_pb2.Value()



import asyncio
from typing import List, TYPE_CHECKING

from .base import BasePrefetcher
from ....helper import get_or_reuse_loop

__all__ = ['HTTPClientPrefetcher', 'WebsocketPrefetcher']

if TYPE_CHECKING:
    from ....types.request import Request


class ClientPrefetcher(BasePrefetcher):
    """Client Prefetcher to be inherited by HTTP / Websocket Prefetchers"""


class HTTPClientPrefetcher(ClientPrefetcher):
    """An async HTTP request handler used in the HTTP Client"""

    def _create_receive_task(self) -> 'asyncio.Task':
        """Create receive task

        :return: asyncio Task
        """
        return get_or_reuse_loop().create_task(self.receive())

    async def receive(self):
        """
        For HTTP Client, there's no task needed for receiving.
        Sleep like there's no tomorrow!
        """
        await asyncio.sleep(1e9)

    def convert_to_message(self, request: 'Request', **kwargs):
        """Convert request to dict for POST request

        :param request: request from client
        :param kwargs: keyword args
        :return: request as dict
        """
        req_dict = request.dict()
        req_dict['exec_endpoint'] = req_dict['header']['exec_endpoint']
        req_dict['data'] = req_dict['data'].get('docs', None)
        return req_dict

    def handle_request(self, request: 'Request') -> 'asyncio.Task':
        """
        For HTTP Client, for each request in the iterator, we send the message (http POST request)
        and add it to the list of tasks which is awaited.

        :param request: current request in the iterator
        :return: asyncio Task for sending message
        """
        return asyncio.create_task(
            self.iolet.send_message(self.convert_to_message(request=request))
        )


class WebsocketPrefetcher(ClientPrefetcher):
    """An async request/response handler used in the Websocket Client (tbd)"""



from abc import ABC, abstractmethod
from typing import Union, List, Tuple, Iterable, Iterator


if False:
    from .document import DocumentArray
    from .traversable import TraversableSequence
    from ..document import Document


class AbstractDocumentArray(ABC):
    """ Abstract class that defines the public interface of DocumentArray classes """

    @abstractmethod
    def get_attributes(self, *args, **kwargs) -> Union[List, List[List]]:
        """
        Return all nonempty values of the fields from all docs this array contains

        :param args: arguments
        :param kwargs: keyword arguments
        """
        ...

    @abstractmethod
    def get_attributes_with_docs(
        self, *args, **kwargs
    ) -> Tuple[Union[List, List[List]], 'DocumentArray']:
        """
        Returns all nonempty values of the fields together with their nonempty docs

        :param args: arguments
        :param kwargs: keyword arguments
        """
        ...

    @abstractmethod
    def traverse(self, *args, **kwargs) -> Iterable['TraversableSequence']:
        """
        Returns an Iterator of :class:``TraversableSequence`` of the leaves when applying the traversal_paths.
        Each :class:``TraversableSequence`` is either the root Documents, a ChunkArray or a MatchArray.

        :param args: arguments
        :param kwargs: keyword arguments
        """
        ...

    @abstractmethod
    def traverse_flat_per_path(
        self, *args, **kwargs
    ) -> Iterable['TraversableSequence']:
        """
        Returns a flattened :class:``TraversableSequence`` per path.

        :param args: arguments
        :param kwargs: keyword arguments
        """
        ...

    @abstractmethod
    def traverse_flat(self, *args, **kwargs) -> 'TraversableSequence':
        """
        Returns a single flattened :class:``TraversableSequence`` with all Documents, that are reachable by traversal
        paths.

        :param args: arguments
        :param kwargs: keyword arguments
        """
        ...

    @abstractmethod
    def match(self, *args, **kwargs) -> None:
        """Computes embedding-based nearest neighbour in another document array for each Document in `self`,
        and store results in the `matches` attribute of each Document.

        :param args: arguments
        :param kwargs: keyword arguments
        """
        ...

    @abstractmethod
    def visualize(self, *args, **kwargs):
        """Visualizes embeddings in a 2D projection with the PCA algorithm. This function requires ``matplotlib``
        installed.

        :param args: arguments
        :param kwargs: keyword arguments
        """
        ...

    @abstractmethod
    def sample(self, *args, **kwargs) -> 'DocumentArray':
        """
        Randomly sample k elements from  the document array without replacement.

        :param args: arguments
        :param kwargs: keyword arguments
        """
        ...

    @abstractmethod
    def shuffle(self, *args, **kwargs) -> 'DocumentArray':
        """
        Randomly shuffle documents within the document array.

        :param args: arguments
        :param kwargs: keyword arguments
        """
        ...

    @abstractmethod
    def __eq__(self):
        ...

    @abstractmethod
    def __len__(self):
        ...

    @abstractmethod
    def __iter__(self) -> Iterator['Document']:
        ...

    @abstractmethod
    def __contains__(self):
        ...

    @abstractmethod
    def __getitem__(self, item: Union[int, str, slice]):
        ...

    @abstractmethod
    def __setitem__(self, key, value: 'Document'):
        ...

    @abstractmethod
    def __delitem__(self, index: Union[int, str, slice]):
        ...

    @abstractmethod
    def extend(self, *args, **kwargs) -> None:
        """
        Extend the document array by appending all the items from the iterable.

        :param args: arguments
        :param kwargs: keyword arguments
        """
        ...

    @abstractmethod
    def append(self, *args, **kwargs):
        """
        Append the Document to the document array.

        :param args: arguments
        :param kwargs: keyword arguments
        """
        ...



import copy
import os
import time
from argparse import Namespace
from typing import Optional, Dict, Union, Set, List

import jina
from .k8slib import kubernetes_deployment, kubernetes_tools
from ..pods import BasePod, ExitFIFO
from ... import __default_executor__
from ...logging.logger import JinaLogger
from ...excepts import RuntimeFailToStart


class K8sPod(BasePod, ExitFIFO):
    """The K8sPod (KubernetesPod)  is used for deployments on Kubernetes."""

    class _K8sDeployment:
        def __init__(
            self,
            name: str,
            head_port_in: int,
            tail_port_out: int,
            head_zmq_identity: bytes,
            version: str,
            shard_id: Optional[int],
            common_args: Union['Namespace', Dict],
            deployment_args: Union['Namespace', Dict],
        ):
            self.name = name
            self.dns_name = kubernetes_deployment.to_dns_name(name)
            self.head_port_in = head_port_in
            self.tail_port_out = tail_port_out
            self.head_zmq_identity = head_zmq_identity
            self.version = version
            self.shard_id = shard_id
            self.common_args = common_args
            self.deployment_args = deployment_args
            self.k8s_namespace = self.common_args.k8s_namespace
            self.num_replicas = getattr(self.deployment_args, 'replicas', 1)

        def _deploy_gateway(self):
            test_pip = os.getenv('JINA_K8S_USE_TEST_PIP') is not None
            image_name = (
                'jinaai/jina:test-pip'
                if test_pip
                else f'jinaai/jina:{self.version}-py38-perf'
            )
            kubernetes_deployment.deploy_service(
                self.dns_name,
                namespace=self.k8s_namespace,
                image_name=image_name,
                container_cmd='["jina"]',
                container_args=f'["gateway", '
                f'"--grpc-data-requests", '
                f'{kubernetes_deployment.get_cli_params(self.common_args, ("pod_role",))}]',
                logger=JinaLogger(f'deploy_{self.name}'),
                replicas=1,
                pull_policy='IfNotPresent',
                port_expose=self.common_args.port_expose,
            )

        @staticmethod
        def _construct_runtime_container_args(
            deployment_args, uses, uses_metas, uses_with_string
        ):
            container_args = (
                f'["executor", '
                f'"--native", '
                f'"--uses", "{uses}", '
                f'"--grpc-data-requests", '
                f'"--runtime-cls", "GRPCDataRuntime", '
                f'"--uses-metas", "{uses_metas}", '
                + uses_with_string
                + f'{kubernetes_deployment.get_cli_params(deployment_args)}]'
            )
            return container_args

        def _deploy_runtime(self):
            image_name = kubernetes_deployment.get_image_name(self.deployment_args.uses)
            init_container_args = kubernetes_deployment.get_init_container_args(
                self.common_args
            )
            uses_metas = kubernetes_deployment.dictionary_to_cli_param(
                {'pea_id': self.shard_id}
            )
            uses_with = kubernetes_deployment.dictionary_to_cli_param(
                self.deployment_args.uses_with
            )
            uses_with_string = f'"--uses-with", "{uses_with}", ' if uses_with else ''
            if image_name == __default_executor__:
                test_pip = os.getenv('JINA_K8S_USE_TEST_PIP') is not None
                image_name = (
                    'jinaai/jina:test-pip'
                    if test_pip
                    else f'jinaai/jina:{self.version}-py38-perf'
                )
                uses = 'BaseExecutor'
            else:
                uses = 'config.yml'
            container_args = self._construct_runtime_container_args(
                self.deployment_args, uses, uses_metas, uses_with_string
            )

            kubernetes_deployment.deploy_service(
                self.dns_name,
                namespace=self.k8s_namespace,
                image_name=image_name,
                container_cmd='["jina"]',
                container_args=container_args,
                logger=JinaLogger(f'deploy_{self.name}'),
                replicas=self.num_replicas,
                pull_policy='IfNotPresent',
                init_container=init_container_args,
                env=self.deployment_args.env,
                custom_resource_dir=getattr(
                    self.common_args, 'k8s_custom_resource_dir', None
                ),
            )

        def wait_start_success(self):
            _timeout = self.common_args.timeout_ready
            if _timeout <= 0:
                _timeout = None
            else:
                _timeout /= 1e3

            from kubernetes import client

            k8s_client = kubernetes_tools.K8sClients().apps_v1

            with JinaLogger(f'waiting_for_{self.name}') as logger:
                logger.info(
                    f'🏝️\n\t\tWaiting for "{self.name}" to be ready, with {self.num_replicas} replicas'
                )
                timeout_ns = 1000000000 * _timeout if _timeout else None
                now = time.time_ns()
                exception_to_raise = None
                while timeout_ns is None or time.time_ns() - now < timeout_ns:
                    try:
                        api_response = k8s_client.read_namespaced_deployment(
                            name=self.dns_name, namespace=self.k8s_namespace
                        )
                        assert api_response.status.replicas == self.num_replicas
                        if (
                            api_response.status.ready_replicas is not None
                            and api_response.status.ready_replicas == self.num_replicas
                        ):
                            logger.success(f' {self.name} has all its replicas ready!!')
                            return
                        else:
                            ready_replicas = api_response.status.ready_replicas or 0
                            logger.info(
                                f'\nNumber of ready replicas {ready_replicas}, waiting for {self.num_replicas - ready_replicas} replicas to be available'
                            )
                            time.sleep(1.0)
                    except client.ApiException as ex:
                        exception_to_raise = ex
                        break
            fail_msg = f' Deployment {self.name} did not start with a timeout of {self.common_args.timeout_ready}'
            if exception_to_raise:
                fail_msg += f': {repr(exception_to_raise)}'
            raise RuntimeFailToStart(fail_msg)

        def start(self):
            with JinaLogger(f'start_{self.name}') as logger:
                logger.info(f'\t\tDeploying "{self.name}"')
                if self.name == 'gateway':
                    self._deploy_gateway()
                else:
                    self._deploy_runtime()
                if not self.common_args.noblock_on_start:
                    self.wait_start_success()
            return self

        def close(self):
            from kubernetes import client

            k8s_client = kubernetes_tools.K8sClients().apps_v1

            with JinaLogger(f'close_{self.name}') as logger:
                try:
                    resp = k8s_client.delete_namespaced_deployment(
                        name=self.dns_name, namespace=self.k8s_namespace
                    )
                    if resp.status == 'Success':
                        logger.success(
                            f' Successful deletion of deployment {self.name}'
                        )
                    else:
                        logger.error(
                            f' Deletion of deployment {self.name} unsuccessful with status {resp.status}'
                        )
                except client.ApiException as exc:
                    logger.error(
                        f' Error deleting deployment {self.name}: {exc.reason} '
                    )

        def __enter__(self):
            return self.start()

        def __exit__(self, exc_type, exc_val, exc_tb):
            self.close()

        def to_node(self):
            return {
                'name': self.dns_name,
                'head_host': f'{self.dns_name}.{self.k8s_namespace}.svc',
                'head_port_in': self.head_port_in,
                'tail_port_out': self.tail_port_out,
                'head_zmq_identity': self.head_zmq_identity,
            }

    def __init__(
        self, args: Union['Namespace', Dict], needs: Optional[Set[str]] = None
    ):
        super().__init__()
        self.args = args
        self.needs = needs or set()
        self.deployment_args = self._parse_args(args)
        self.version = self._get_base_executor_version()

        self.fixed_head_port_in = 8081
        self.fixed_tail_port_out = 8082
        self.k8s_head_deployment = None
        self.k8s_tail_deployment = None
        if self.deployment_args['head_deployment'] is not None:
            name = f'{self.name}-head'
            self.k8s_head_deployment = self._K8sDeployment(
                name=name,
                head_port_in=self.fixed_head_port_in,
                tail_port_out=self.fixed_tail_port_out,
                head_zmq_identity=self.head_zmq_identity,
                version=self.version,
                shard_id=None,
                common_args=self.args,
                deployment_args=self.deployment_args['head_deployment'],
            )
        if self.deployment_args['tail_deployment'] is not None:
            name = f'{self.name}-tail'
            self.k8s_tail_deployment = self._K8sDeployment(
                name=name,
                head_port_in=self.fixed_head_port_in,
                tail_port_out=self.fixed_tail_port_out,
                head_zmq_identity=self.head_zmq_identity,
                version=self.version,
                shard_id=None,
                common_args=self.args,
                deployment_args=self.deployment_args['tail_deployment'],
            )

        self.k8s_deployments = []
        for i, args in enumerate(self.deployment_args['deployments']):
            name = (
                f'{self.name}-{i}'
                if len(self.deployment_args['deployments']) > 1
                else f'{self.name}'
            )
            self.k8s_deployments.append(
                self._K8sDeployment(
                    name=name,
                    head_port_in=self.fixed_head_port_in,
                    tail_port_out=self.fixed_tail_port_out,
                    head_zmq_identity=self.head_zmq_identity,
                    version=self.version,
                    shard_id=i,
                    common_args=self.args,
                    deployment_args=args,
                )
            )

    def _parse_args(
        self, args: Namespace
    ) -> Dict[str, Optional[Union[List[Namespace], Namespace]]]:
        return self._parse_deployment_args(args)

    def _parse_deployment_args(self, args):
        parsed_args = {
            'head_deployment': None,
            'tail_deployment': None,
            'deployments': [],
        }
        shards = getattr(args, 'shards', 1)
        replicas = getattr(args, 'replicas', 1)
        uses_before = getattr(args, 'uses_before', None)
        if shards > 1 or (len(self.needs) > 1 and replicas > 1) or uses_before:
            # reasons to separate head and tail from peas is that they
            # can be deducted based on the previous and next pods
            parsed_args['head_deployment'] = copy.copy(args)
            parsed_args['head_deployment'].replicas = 1
            parsed_args['head_deployment'].uses = (
                args.uses_before or __default_executor__
            )
        if shards > 1 or getattr(args, 'uses_after', None):
            parsed_args['tail_deployment'] = copy.copy(args)
            parsed_args['tail_deployment'].replicas = 1
            parsed_args['tail_deployment'].uses = (
                args.uses_after or __default_executor__
            )

        parsed_args['deployments'] = [args] * shards
        return parsed_args

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        super().__exit__(exc_type, exc_val, exc_tb)
        self.join()

    @property
    def port_expose(self) -> int:
        """Not implemented"""
        raise NotImplementedError

    @property
    def host(self) -> str:
        """Currently, when deploying on Kubernetes, Jina does not expose a public host.
        Instead Jina sends requests via port-forward and runs the requests against localhost.

        :return: localhost
        """
        return 'localhost'

    def start(self) -> 'K8sPod':
        """Deploy the kubernetes pods via k8s Deployment and k8s Service.

        :return: self
        """
        with JinaLogger(f'start_{self.name}') as logger:
            logger.info(f'🏝️\tCreate deployments for "{self.name}"')
            if self.k8s_head_deployment is not None:
                self.enter_context(self.k8s_head_deployment)
            for k8s_deployment in self.k8s_deployments:
                self.enter_context(k8s_deployment)
            if self.k8s_tail_deployment is not None:
                self.enter_context(self.k8s_tail_deployment)
        return self

    def wait_start_success(self):
        """Not implemented. It should wait until the deployment is up and running"""
        if not self.args.noblock_on_start:
            raise ValueError(
                f'{self.wait_start_success!r} should only be called when `noblock_on_start` is set to True'
            )
        try:
            if self.k8s_head_deployment is not None:
                self.k8s_head_deployment.wait_start_success()
            for p in self.k8s_deployments:
                p.wait_start_success()
            if self.k8s_tail_deployment is not None:
                self.k8s_tail_deployment.wait_start_success()
        except:
            self.close()
            raise

    def join(self):
        """Not needed. The context managers will manage the proper deletion"""
        pass

    def update_pea_args(self):
        """
        Regenerate deployment args
        """
        self.deployment_args = self._parse_args(self.args)

    @property
    def head_args(self) -> Namespace:
        """Head args of the pod.

        :return: namespace
        """
        return self.args

    @property
    def tail_args(self) -> Namespace:
        """Tail args of the pod

        :return: namespace
        """
        return self.args

    @property
    def num_peas(self) -> int:
        """Number of peas. Currently unused.

        :return: number of peas
        """
        return sum(
            [
                self.k8s_head_deployment.num_replicas
                if self.k8s_head_deployment is not None
                else 0
            ]
            + [
                self.k8s_tail_deployment.num_replicas
                if self.k8s_tail_deployment is not None
                else 0
            ]
            + [k8s_deployment.num_replicas for k8s_deployment in self.k8s_deployments]
        )

    @property
    def head_zmq_identity(self) -> bytes:
        """zmq identity is not needed for k8s deployment

        :return: zmq identity
        """
        return b''

    @property
    def deployments(self) -> List[Dict]:
        """Deployment information which describes the interface of the pod.

        :return: list of dictionaries defining the attributes used by the routing table
        """
        res = []

        if self.args.name == 'gateway':
            res.append(self.k8s_deployments[0].to_node())
        else:
            if self.k8s_head_deployment:
                res.append(self.k8s_head_deployment.to_node())
            res.extend([_.to_node() for _ in self.k8s_deployments])
            if self.k8s_tail_deployment:
                res.append(self.k8s_tail_deployment.to_node())
        return res

    def _get_base_executor_version(self):
        import requests

        url = 'https://registry.hub.docker.com/v1/repositories/jinaai/jina/tags'
        tags = requests.get(url).json()
        name_set = {tag['name'] for tag in tags}
        if jina.__version__ in name_set:
            return jina.__version__
        else:
            return 'master'

    @property
    def _mermaid_str(self) -> List[str]:
        """String that will be used to represent the Pod graphically when `Flow.plot()` is invoked


        .. # noqa: DAR201
        """
        mermaid_graph = []
        if self.name != 'gateway':
            mermaid_graph = [f'subgraph {self.name};\n', f'direction LR;\n']

            num_replicas = getattr(self.args, 'replicas', 1)
            num_shards = getattr(self.args, 'shards', 1)
            uses = self.args.uses
            if num_shards > 1:
                shard_names = [
                    f'{args.name}/shard-{i}'
                    for i, args in enumerate(self.deployment_args['deployments'])
                ]
                for shard_name in shard_names:
                    shard_mermaid_graph = [
                        f'subgraph {shard_name}\n',
                        f'direction TB;\n',
                    ]
                    for replica_id in range(num_replicas):
                        shard_mermaid_graph.append(
                            f'{shard_name}/replica-{replica_id}[{uses}]\n'
                        )
                    shard_mermaid_graph.append(f'end\n')
                    mermaid_graph.extend(shard_mermaid_graph)
                head_name = f'{self.name}/head'
                tail_name = f'{self.name}/tail'
                head_to_show = self.args.uses_before
                if head_to_show is None or head_to_show == __default_executor__:
                    head_to_show = head_name
                tail_to_show = self.args.uses_after
                if tail_to_show is None or tail_to_show == __default_executor__:
                    tail_to_show = tail_name
                if head_name:
                    for shard_name in shard_names:
                        mermaid_graph.append(
                            f'{head_name}[{head_to_show}]:::HEADTAIL --> {shard_name}[{uses}];'
                        )

                if tail_name:
                    for shard_name in shard_names:
                        mermaid_graph.append(
                            f'{shard_name}[{uses}] --> {tail_name}[{tail_to_show}]:::HEADTAIL;'
                        )
            else:
                for replica_id in range(num_replicas):
                    mermaid_graph.append(f'{self.name}/replica-{replica_id}[{uses}];')

            mermaid_graph.append(f'end;')
        return mermaid_graph



import os
import asyncio

from jina import __default_host__

from .....importer import ImportExtensions
from ...zmq.asyncio import AsyncNewLoopRuntime
from .app import get_fastapi_app

__all__ = ['WebSocketRuntime']


class WebSocketRuntime(AsyncNewLoopRuntime):
    """Runtime for Websocket interface."""

    async def async_setup(self):
        """
        The async method setup the runtime.

        Setup the uvicorn server.
        """
        with ImportExtensions(required=True):
            from uvicorn import Config, Server

        class UviServer(Server):
            """The uvicorn server."""

            async def setup(self, sockets=None):
                """
                Setup uvicorn server.

                :param sockets: sockets of server.
                """
                config = self.config
                if not config.loaded:
                    config.load()
                self.lifespan = config.lifespan_class(config)
                self.install_signal_handlers()
                await self.startup(sockets=sockets)
                if self.should_exit:
                    return

            async def serve(self, **kwargs):
                """
                Start the server.

                :param kwargs: keyword arguments
                """
                await self.main_loop()

        from .....helper import extend_rest_interface

        uvicorn_kwargs = self.args.uvicorn_kwargs or {}
        self._server = UviServer(
            config=Config(
                app=extend_rest_interface(get_fastapi_app(self.args, self.logger)),
                host=__default_host__,
                port=self.args.port_expose,
                ws_max_size=1024 * 1024 * 1024,
                log_level=os.getenv('JINA_LOG_LEVEL', 'error').lower(),
                **uvicorn_kwargs
            )
        )
        await self._server.setup()

    async def async_run_forever(self):
        """Running method of ther server."""
        await self._server.serve()

    async def _wait_for_cancel(self):
        """Do NOT override this method when inheriting from :class:`GatewayPea`"""
        # handle terminate signals
        while not self.is_cancel.is_set() and not self._server.should_exit:
            await asyncio.sleep(0.1)

        await self.async_cancel()

    async def async_cancel(self):
        """Stop the server."""
        self._server.should_exit = True
        await self._server.shutdown()



import asyncio
import argparse

from typing import Dict, List, Union, TYPE_CHECKING

from .base import BasePrefetcher
from ....types.message import Message
from ....helper import get_or_reuse_loop

__all__ = ['ZmqGatewayPrefetcher', 'GrpcGatewayPrefetcher']

if TYPE_CHECKING:
    from ...grpc import Grpclet
    from ...zmq import AsyncZmqlet
    from ....types.request import Request, Response


class GatewayPrefetcher(BasePrefetcher):
    """Gateway Prefetcher to be inherited by ZMQ / GRPC Prefetchers"""

    def __init__(
        self,
        args: argparse.Namespace,
        iolet: Union['AsyncZmqlet', 'Grpclet'],
    ):
        super().__init__(args, iolet)
        self.request_buffer: Dict[str, asyncio.Future[Message]] = dict()
        self.Call = self.send  # Used in grpc servicer

    def convert_to_message(self, request: 'Request'):
        """Convert a `Request` to a `Message` to be sent from gateway

        :param request: request from iterator
        :return: Message from request
        """
        return Message(None, request, 'gateway', **vars(self.args))

    def handle_request(self, request: 'Request') -> 'asyncio.Future':
        """
        For ZMQ & GRPC data requests, for each request in the iterator, we send the `Message` using
        `iolet.send_message()` and add {<request-id>: <an-empty-future>} to the message buffer.
        This empty future is used to track the `result` of this request during `receive`

        :param request: current request in the iterator
        :return: asyncio Future for sending message
        """
        future = get_or_reuse_loop().create_future()
        self.request_buffer[request.request_id] = future
        asyncio.create_task(
            self.iolet.send_message(self.convert_to_message(request=request))
        )
        return future

    async def receive(self):
        """Await messages back from Executors and process them in the message buffer"""
        try:
            while True:
                response = await self.iolet.recv_message(callback=lambda x: x.response)
                # during shutdown the socket will return None
                if response is None:
                    break

                self.handle_response(response)
        except asyncio.CancelledError:
            raise
        finally:
            for future in self.request_buffer.values():
                future.cancel(
                    f'{self.__class__.__name__} closed, all outstanding requests canceled'
                )
            self.request_buffer.clear()

    def handle_response(self, response: 'Response'):
        """
        Set result of each response received from Executors in the request buffer

        :param response: message received during `iolet.recv_message`
        """
        if response.request_id in self.request_buffer:
            future = self.request_buffer.pop(response.request_id)
            future.set_result(response)
        else:
            self.logger.warning(
                f'Discarding unexpected response with request id {response.request_id}'
            )


class ZmqGatewayPrefetcher(GatewayPrefetcher):
    """An async zmq request handler used in the Gateway"""

    def _create_receive_task(self):
        """Start a receive task that starts the GRPC server & awaits termination.

        :return: asyncio Task
        """
        return get_or_reuse_loop().create_task(self.receive())


class GrpcGatewayPrefetcher(GatewayPrefetcher):
    """An async grpc request handler used in the Gateway"""

    def __init__(self, args: argparse.Namespace, iolet: 'Grpclet'):
        super().__init__(args, iolet)
        self.iolet.callback = lambda response: self.handle_response(response.request)

    async def handle_response(self, response: 'Response'):
        """
        Async version of parents handle_response function

        :param response: message received from grpclet callback
        """
        super().handle_response(response)

    def _create_receive_task(self) -> 'asyncio.Task':
        """Start a receive task that starts the GRPC server & awaits termination.

        :return: asyncio Task
        """
        return get_or_reuse_loop().create_task(self.iolet.start())



import argparse
import os
import sys
import time
import warnings
from typing import Union, Optional, Dict, TYPE_CHECKING
from pathlib import Path
from platform import uname

from ..zmq.base import ZMQRuntime
from ...zmq import Zmqlet
from .... import __docker_host__
from .helper import get_docker_network, get_gpu_device_requests
from ....excepts import BadImageNameError, DockerVersionError
from ...zmq import send_ctrl_message
from ....helper import ArgNamespace, slugify, random_name
from ....enums import SocketType

if TYPE_CHECKING:
    import multiprocessing
    import threading
    from ....logging.logger import JinaLogger


class ContainerRuntime(ZMQRuntime):
    """Runtime procedure for container."""

    def __init__(self, args: 'argparse.Namespace', **kwargs):
        super().__init__(args, **kwargs)
        self.ctrl_addr = self.get_control_address(
            args.host,
            args.port_ctrl,
            docker_kwargs=self.args.docker_kwargs,
        )
        if (
            self.args.docker_kwargs
            and 'extra_hosts' in self.args.docker_kwargs
            and __docker_host__ in self.args.docker_kwargs['extra_hosts']
        ):
            self.args.docker_kwargs.pop('extra_hosts')
        self._set_network_for_dind_linux()
        self._docker_run()
        while self._is_container_alive and not self.is_ready:
            time.sleep(1)
        # two cases to reach here: 1. is_ready, 2. container is dead
        if not self._is_container_alive:
            # replay it to see the log
            self._docker_run(replay=True)
            raise Exception(
                'the container fails to start, check the arguments or entrypoint'
            )

    def teardown(self):
        """Stop the container."""
        self._container.stop()
        super().teardown()

    def _stream_logs(self):
        for line in self._container.logs(stream=True):
            self.logger.info(line.strip().decode())

    def run_forever(self):
        """Stream the logs while running."""
        self._stream_logs()

    def _set_network_for_dind_linux(self):
        import docker

        # recompute the control_addr, do not assign client, since this would be an expensive object to
        # copy in the new process generated
        client = docker.from_env()

        # Related to potential docker-in-docker communication. If `ContainerPea` lives already inside a container.
        # it will need to communicate using the `bridge` network.
        self._net_mode = None

        # In WSL, we need to set ports explicitly
        if sys.platform in ('linux', 'linux2') and 'microsoft' not in uname().release:
            self._net_mode = 'host'
            try:
                bridge_network = client.networks.get('bridge')
                if bridge_network:
                    self.ctrl_addr, _ = Zmqlet.get_ctrl_address(
                        bridge_network.attrs['IPAM']['Config'][0]['Gateway'],
                        self.args.port_ctrl,
                        self.args.ctrl_with_ipc,
                    )
            except Exception as ex:
                self.logger.warning(
                    f'Unable to set control address from "bridge" network: {ex!r}'
                    f' Control address set to {self.ctrl_addr}'
                )
        client.close()

    def _docker_run(self, replay: bool = False):
        # important to notice, that client is not assigned as instance member to avoid potential
        # heavy copy into new process memory space
        import docker

        client = docker.from_env()

        docker_version = client.version().get('Version')
        if not docker_version:
            raise DockerVersionError('docker version can not be resolved')

        docker_version = tuple(docker_version.split('.'))
        # docker daemon versions below 20.0x do not support "host.docker.internal:host-gateway"
        if docker_version < ('20',):
            raise DockerVersionError(
                f'docker version {".".join(docker_version)} is below 20.0.0 and does not '
                f'support "host.docker.internal:host-gateway" : https://github.com/docker/cli/issues/2664'
            )

        if self.args.uses.startswith('docker://'):
            uses_img = self.args.uses.replace('docker://', '')
            self.logger.debug(f'will use Docker image: {uses_img}')
        else:
            warnings.warn(
                f'you are using legacy image format {self.args.uses}, this may create some ambiguity. '
                f'please use the new format: "--uses docker://{self.args.uses}"'
            )
            uses_img = self.args.uses

        # the image arg should be ignored otherwise it keeps using ContainerPea in the container
        # basically all args in BasePea-docker arg group should be ignored.
        # this prevent setting containerPea twice
        from ....parsers import set_pea_parser

        self.args.runs_in_docker = True
        self.args.native = True
        non_defaults = ArgNamespace.get_non_defaults_args(
            self.args,
            set_pea_parser(),
            taboo={
                'uses',
                'entrypoint',
                'volumes',
                'pull_latest',
                'runtime_cls',
                'docker_kwargs',
                'gpus',
            },
        )
        img_not_found = False

        try:
            client.images.get(uses_img)
        except docker.errors.ImageNotFound:
            self.logger.error(f'can not find local image: {uses_img}')
            img_not_found = True

        if self.args.pull_latest or img_not_found:
            self.logger.warning(
                f'pulling {uses_img}, this could take a while. if you encounter '
                f'timeout error due to pulling takes to long, then please set '
                f'"timeout-ready" to a larger value.'
            )
            try:
                client.images.pull(uses_img)
                img_not_found = False
            except docker.errors.NotFound:
                img_not_found = True
                self.logger.error(f'can not find remote image: {uses_img}')

        if img_not_found:
            raise BadImageNameError(
                f'image: {uses_img} can not be found local & remote.'
            )

        _volumes = {}
        if self.args.volumes:
            for p in self.args.volumes:
                paths = p.split(':')
                local_path = paths[0]
                Path(os.path.abspath(local_path)).mkdir(parents=True, exist_ok=True)
                if len(paths) == 2:
                    container_path = paths[1]
                else:
                    container_path = '/' + os.path.basename(p)
                _volumes[os.path.abspath(local_path)] = {
                    'bind': container_path,
                    'mode': 'rw',
                }

        device_requests = []
        if self.args.gpus:
            device_requests = get_gpu_device_requests(self.args.gpus)
            del self.args.gpus

        _expose_port = [self.args.port_ctrl]
        if self.args.socket_in.is_bind:
            _expose_port.append(self.args.port_in)
        if self.args.socket_out.is_bind:
            _expose_port.append(self.args.port_out)

        _args = ArgNamespace.kwargs2list(non_defaults)
        ports = {f'{v}/tcp': v for v in _expose_port} if not self._net_mode else None

        docker_kwargs = self.args.docker_kwargs or {}
        self._container = client.containers.run(
            uses_img,
            _args,
            detach=True,
            auto_remove=True,
            ports=ports,
            name=slugify(f'{self.name}/{random_name()}'),
            volumes=_volumes,
            network_mode=self._net_mode,
            entrypoint=self.args.entrypoint,
            extra_hosts={__docker_host__: 'host-gateway'},
            device_requests=device_requests,
            **docker_kwargs,
        )

        if replay:
            # when replay is on, it means last time it fails to start
            # therefore we know the loop below wont block the main process
            self._stream_logs()

        client.close()

    @property
    def status(self):
        """
        Send get status control message.

        :return: control message.
        """
        return send_ctrl_message(
            self.ctrl_addr, 'STATUS', timeout=self.args.timeout_ctrl
        )

    @property
    def is_ready(self) -> bool:
        """
        Check if status is ready.

        :return: True if status is ready else False.
        """
        status = self.status
        return status and status.is_ready

    @property
    def _is_container_alive(self) -> bool:
        import docker.errors

        try:
            self._container.reload()
        except docker.errors.NotFound:
            return False
        return True

    # Static methods used by the Pea to communicate with the `Runtime` in the separate process
    @staticmethod
    def wait_for_ready_or_shutdown(
        timeout: Optional[float],
        ready_or_shutdown_event: Union['multiprocessing.Event', 'threading.Event'],
        **kwargs,
    ):
        """
        Check if the runtime has successfully started

        :param timeout: The time to wait before readiness or failure is determined
        :param ready_or_shutdown_event: the multiprocessing event to detect if the process failed or succeeded
        :param kwargs: extra keyword arguments

        :return: True if is ready or it needs to be shutdown
        """
        return ready_or_shutdown_event.wait(timeout)

    @staticmethod
    def _retry_control_message(
        ctrl_address: str,
        timeout_ctrl: int,
        command: str,
        num_retry: int,
        logger: 'JinaLogger',
    ):
        from ...zmq import send_ctrl_message

        for retry in range(1, num_retry + 1):
            logger.debug(f'Sending {command} command for the {retry}th time')
            try:
                send_ctrl_message(
                    ctrl_address,
                    command,
                    timeout=timeout_ctrl,
                    raise_exception=True,
                )
                break
            except Exception as ex:
                logger.warning(f'{ex!r}')
                if retry == num_retry:
                    raise ex

    @staticmethod
    def cancel(
        control_address: str,
        timeout_ctrl: int,
        socket_in_type: 'SocketType',
        skip_deactivate: bool,
        logger: 'JinaLogger',
        **kwargs,
    ):
        """
        Check if the runtime has successfully started

        :param control_address: the address where the control message needs to be sent
        :param timeout_ctrl: the timeout to wait for control messages to be processed
        :param socket_in_type: the type of input socket, needed to know if is a dealer
        :param skip_deactivate: flag to tell if deactivate signal may be missed.
            This is important when you want to independently kill a Runtime
        :param logger: the JinaLogger to log messages
        :param kwargs: extra keyword arguments
        """
        if not skip_deactivate and socket_in_type == SocketType.DEALER_CONNECT:
            ContainerRuntime._retry_control_message(
                ctrl_address=control_address,
                timeout_ctrl=timeout_ctrl,
                command='DEACTIVATE',
                num_retry=3,
                logger=logger,
            )
        ContainerRuntime._retry_control_message(
            ctrl_address=control_address,
            timeout_ctrl=timeout_ctrl,
            command='TERMINATE',
            num_retry=3,
            logger=logger,
        )

    @staticmethod
    def activate(
        control_address: str,
        timeout_ctrl: int,
        socket_in_type: 'SocketType',
        logger: 'JinaLogger',
        **kwargs,
    ):
        """
        Check if the runtime has successfully started

        :param control_address: the address where the control message needs to be sent
        :param timeout_ctrl: the timeout to wait for control messages to be processed
        :param socket_in_type: the type of input socket, needed to know if is a dealer
        :param logger: the JinaLogger to log messages
        :param kwargs: extra keyword arguments
        """
        if socket_in_type == SocketType.DEALER_CONNECT:
            ContainerRuntime._retry_control_message(
                ctrl_address=control_address,
                timeout_ctrl=timeout_ctrl,
                command='ACTIVATE',
                num_retry=3,
                logger=logger,
            )

    @staticmethod
    def get_control_address(
        host: str,
        port: str,
        docker_kwargs: Optional[Dict],
        **kwargs,
    ):
        """
        Get the control address for a runtime with a given host and port

        :param host: the host where the runtime works
        :param port: the control port where the runtime listens
        :param docker_kwargs: the extra docker kwargs from which maybe extract extra hosts
        :param kwargs: extra keyword arguments
        :return: The corresponding control address
        """
        import docker

        client = docker.from_env()
        network = get_docker_network(client)

        if (
            docker_kwargs
            and 'extra_hosts' in docker_kwargs
            and __docker_host__ in docker_kwargs['extra_hosts']
        ):
            ctrl_host = __docker_host__
        elif network:
            # If the caller is already in a docker network, replace ctrl-host with network gateway
            try:
                ctrl_host = client.networks.get(network).attrs['IPAM']['Config'][0][
                    'Gateway'
                ]
            except Exception:
                ctrl_host = __docker_host__
        else:
            ctrl_host = host

        return Zmqlet.get_ctrl_address(ctrl_host, port, False)[0]



import argparse
import asyncio
from abc import ABC, abstractmethod
from typing import AsyncGenerator, List, Union, TYPE_CHECKING

from ...grpc import Grpclet
from ....helper import typename
from ....logging.logger import JinaLogger
from ....types.message import Message

__all__ = ['BasePrefetcher']

if TYPE_CHECKING:
    from ...zmq import AsyncZmqlet
    from ....types.request import Request
    from ....clients.base.http import HTTPClientlet


class BasePrefetcher(ABC):
    """An base async request/response handler"""

    def __init__(
        self,
        args: argparse.Namespace,
        iolet: Union['AsyncZmqlet', 'Grpclet', 'HTTPClientlet'],
    ):
        """
        :param args: args from CLI
        :param iolet: One of AsyncZmqlet or Grpclet. Used for sending/receiving data to/from the Flow
        """
        self.args = args
        self.iolet = iolet
        self.receive_task = self._create_receive_task()
        self.logger = JinaLogger(self.__class__.__name__, **vars(args))

    @abstractmethod
    def _create_receive_task(self) -> asyncio.Task:
        """Start a receive task to be running in the background

        .. # noqa: DAR202
        :return: asyncio Task
        """
        ...

    @abstractmethod
    async def receive(self):
        """Implement `receive` logic for prefetcher

        .. # noqa: DAR202
        """
        ...

    @abstractmethod
    def handle_request(
        self, request: 'Request'
    ) -> Union['asyncio.Task', 'asyncio.Future']:
        """Handle each request in the iterator

        :param request: current request in the iterator
        """
        ...

    async def close(self):
        """
        Stop receiving messages
        """
        self.receive_task.cancel()

    async def send(self, request_iterator, *args) -> AsyncGenerator[None, Message]:
        """
        Async call to receive Requests and build them into Messages.

        :param request_iterator: iterator of requests.
        :param args: additional arguments
        :yield: message
        """
        self.args: argparse.Namespace
        self.iolet: Union['AsyncZmqlet', 'Grpclet']
        self.logger: JinaLogger

        if self.receive_task.done():
            raise RuntimeError(
                'PrefetchCaller receive task not running, can not send messages'
            )

        async def prefetch_req(
            num_req: int, fetch_to: List[Union['asyncio.Task', 'asyncio.Future']]
        ):
            """
            Fetch and send request.

            :param num_req: number of requests
            :param fetch_to: the task list storing requests
            :return: False if append task to :param:`fetch_to` else False
            """
            for _ in range(num_req):
                try:
                    if hasattr(request_iterator, '__anext__'):
                        request = await request_iterator.__anext__()
                    elif hasattr(request_iterator, '__next__'):
                        request = next(request_iterator)
                    else:
                        raise TypeError(
                            f'{typename(request_iterator)} does not have `__anext__` or `__next__`'
                        )

                    fetch_to.append(self.handle_request(request=request))
                except (StopIteration, StopAsyncIteration):
                    return True
            return False

        prefetch_task = []
        is_req_empty = await prefetch_req(self.args.prefetch, prefetch_task)
        if is_req_empty and not prefetch_task:
            self.logger.error(
                'receive an empty stream from the client! '
                'please check your client\'s inputs, '
                'you can use "Client.check_input(inputs)"'
            )
            return

        # the total num requests < self.args.prefetch
        if is_req_empty:
            for r in asyncio.as_completed(prefetch_task):
                yield await r
        else:
            # if there are left over (`else` clause above is unnecessary for code but for better readability)
            onrecv_task = []
            # the following code "interleaves" prefetch_task and onrecv_task, when one dries, it switches to the other
            while prefetch_task:
                if self.logger.debug_enabled:
                    self.logger.debug(
                        f'send: {self.iolet.msg_sent} '
                        f'recv: {self.iolet.msg_recv} '
                        f'pending: {self.iolet.msg_sent - self.iolet.msg_recv}'
                    )
                onrecv_task.clear()
                for r in asyncio.as_completed(prefetch_task):
                    yield await r
                    if not is_req_empty:
                        is_req_empty = await prefetch_req(
                            self.args.prefetch_on_recv, onrecv_task
                        )

                # this list dries, clear it and feed it with on_recv_task
                prefetch_task.clear()
                prefetch_task = [j for j in onrecv_task]



from abc import abstractmethod
from typing import Iterable, Sequence, TYPE_CHECKING, Optional, Generator

if TYPE_CHECKING:
    from .document import DocumentArray
    from ..document import Document


def _check_traversal_path_type(tp):
    if not (
        tp
        and not isinstance(tp, str)
        and isinstance(tp, Sequence)
        and all(isinstance(p, str) for p in tp)
    ):
        raise TypeError('`traversal_paths` needs to be `Sequence[str]`')


class TraversableSequence:
    """
    A mixin used for traversing a `Sequence[Traversable]`.
    """

    def traverse(
        self, traversal_paths: Sequence[str]
    ) -> Iterable['TraversableSequence']:
        """
        Return an Iterator of :class:``TraversableSequence`` of the leaves when applying the traversal_paths.
        Each :class:``TraversableSequence`` is either the root Documents, a ChunkArray or a MatchArray.

        :param traversal_paths: a list of string that represents the traversal path
        :yield: :class:``TraversableSequence`` of the leaves when applying the traversal_paths.

        Example on ``traversal_paths``:

            - [`r`]: docs in this TraversableSequence
            - [`m`]: all match-documents at adjacency 1
            - [`c`]: all child-documents at granularity 1
            - [`cc`]: all child-documents at granularity 2
            - [`mm`]: all match-documents at adjacency 2
            - [`cm`]: all match-document at adjacency 1 and granularity 1
            - [`r`, `c`]: docs in this TraversableSequence and all child-documents at granularity 1

        """
        _check_traversal_path_type(traversal_paths)

        for p in traversal_paths:
            yield from self._traverse(self, p)

    @staticmethod
    def _traverse(docs: 'TraversableSequence', path: str):
        if path:
            loc = path[0]
            if loc == 'r':
                yield from TraversableSequence._traverse(docs, path[1:])
            elif loc == 'm':
                for d in docs:
                    yield from TraversableSequence._traverse(d.matches, path[1:])
            elif loc == 'c':
                for d in docs:
                    yield from TraversableSequence._traverse(d.chunks, path[1:])
            else:
                raise ValueError(
                    f'`path`:{loc} is invalid, must be one of `c`, `r`, `m`'
                )
        else:
            yield docs

    def traverse_flat_per_path(self, traversal_paths: Sequence[str]):
        """
        Returns a flattened :class:``TraversableSequence`` per path in :param:``traversal_paths``
        with all Documents, that are reached by the path.

        :param traversal_paths: a list of string that represents the traversal path
        :yield: :class:``TraversableSequence`` containing the document of all leaves per path.
        """
        _check_traversal_path_type(traversal_paths)

        for p in traversal_paths:
            yield self._flatten(self._traverse(self, p))

    def traverse_flat(self, traversal_paths: Sequence[str]) -> Iterable['Document']:
        """
        Returns a single flattened :class:``TraversableSequence`` with all Documents, that are reached
        via the :param:``traversal_paths``.

        .. warning::
            When defining the :param:``traversal_paths`` with multiple paths, the returned
            :class:``Documents`` are determined at once and not on the fly. This is a different
            behavior then in :method:``traverse`` and :method:``traverse_flattened_per_path``!

        :param traversal_paths: a list of string that represents the traversal path
        :return: a single :class:``TraversableSequence`` containing the document of all leaves when applying the traversal_paths.
        """
        _check_traversal_path_type(traversal_paths)
        if len(traversal_paths) == 1 and traversal_paths[0] == 'r':
            return self

        leaves = self.traverse(traversal_paths)
        return self._flatten(leaves)

    def batch(
        self,
        batch_size: int,
        traversal_paths: Sequence[str] = None,
        require_attr: Optional[str] = None,
    ) -> Generator['DocumentArray', None, None]:
        """
        Creates a `Generator` that yields `DocumentArray` of size `batch_size` until `docs` is fully traversed along
        the `traversal_path`. The None `docs` are filtered out and optionally the `docs` can be filtered by checking for
        the existence of a `Document` attribute.
        Note, that the last batch might be smaller than `batch_size`.

        :param traversal_paths: Specifies along which "axis" the document shall be traversed. (defaults to ['r'])
        :param batch_size: Size of each generated batch (except the last one, which might be smaller, default: 32)
        :param require_attr: Optionally, you can filter out docs which don't have this attribute
        :yield: a Generator of `DocumentArray`, each in the length of `batch_size`
        """

        if not (isinstance(batch_size, int) and batch_size > 0):
            raise ValueError('`batch_size` should be a positive integer')

        if traversal_paths:
            _check_traversal_path_type(traversal_paths)
            docs = self.traverse_flat(traversal_paths)
        else:
            docs = self

        from .document import DocumentArray

        _batch = DocumentArray()
        for d in docs:
            # For array-valued attributes we need to compare to None
            if require_attr in ['embedding', 'blob']:
                if getattr(d, require_attr) is not None:
                    _batch.append(d)
            elif require_attr is not None:
                if getattr(d, require_attr):
                    _batch.append(d)
            else:
                _batch.append(d)

            if len(_batch) == batch_size:
                yield _batch
                _batch = DocumentArray()

        if _batch:
            yield _batch

    @staticmethod
    @abstractmethod
    def _flatten(sequence):
        ...



"""Module for helper functions for clients."""
from typing import Tuple

from ...enums import DataInputType
from ...excepts import BadDocType, BadRequestType
from ...types.document import Document
from ...types.request import Request


def _new_data_request_from_batch(
    _kwargs, batch, data_type, endpoint, target, parameters
):
    req = _new_data_request(endpoint, target, parameters)

    # add docs, groundtruths fields
    try:
        _add_docs_groundtruths(req, batch, data_type, _kwargs)
    except Exception as ex:
        raise BadRequestType(
            f'error when building {req.request_type} from {batch}'
        ) from ex

    return req


def _new_data_request(endpoint, target, parameters):
    req = Request()
    req = req.as_typed_request('data')

    # set up header
    if endpoint:
        req.header.exec_endpoint = endpoint
    if target:
        req.header.target_peapod = target
    # add parameters field
    if parameters:
        req.parameters = parameters
    return req


def _new_doc_from_data(
    data, data_type: DataInputType, **kwargs
) -> Tuple['Document', 'DataInputType']:
    def _build_doc_from_content():
        return Document(content=data, **kwargs), DataInputType.CONTENT

    if data_type == DataInputType.AUTO or data_type == DataInputType.DOCUMENT:
        if isinstance(data, Document):
            # if incoming is already primitive type Document, then all good, best practice!
            return data, DataInputType.DOCUMENT
        try:
            d = Document(data, **kwargs)
            return d, DataInputType.DOCUMENT
        except BadDocType:
            # AUTO has a fallback, now reconsider it as content
            if data_type == DataInputType.AUTO:
                return _build_doc_from_content()
            else:
                raise
    elif data_type == DataInputType.CONTENT:
        return _build_doc_from_content()


def _add_docs_groundtruths(req, batch, data_type, _kwargs):
    for content in batch:
        if isinstance(content, tuple) and len(content) == 2:
            # content comes in pair,  will take the first as the input and the second as the ground truth

            # note how data_type is cached
            d, data_type = _new_doc_from_data(content[0], data_type, **_kwargs)
            gt, _ = _new_doc_from_data(content[1], data_type, **_kwargs)
            req.docs.append(d)
            req.groundtruths.append(gt)
        else:
            d, data_type = _new_doc_from_data(content, data_type, **_kwargs)
            req.docs.append(d)


def _add_control_propagate(req, kwargs):
    from ...proto import jina_pb2

    extra_kwargs = kwargs[
        'extra_kwargs'
    ]  #: control command and args are stored inside extra_kwargs
    _available_commands = dict(
        jina_pb2.RequestProto.ControlRequestProto.DESCRIPTOR.enum_values_by_name
    )

    if 'command' in extra_kwargs:
        command = extra_kwargs['command']
    else:
        raise BadRequestType(
            'sending ControlRequest from Client must contain the field `command`'
        )

    if command in _available_commands:
        req.control.command = getattr(
            jina_pb2.RequestProto.ControlRequestProto, command
        )
    else:
        raise ValueError(
            f'command "{command}" is not supported, must be one of {_available_commands}'
        )



from typing import Callable, Dict, Optional
from contextlib import nullcontext, AsyncExitStack

from ..base import BaseClient, InputType
from ..helper import callback_exec
from ...excepts import BadClient
from ...importer import ImportExtensions
from ...logging.profile import ProgressBar
from ...types.request import Request
from ...peapods.runtimes.prefetch.client import HTTPClientPrefetcher


class HTTPClientlet:
    """HTTP Client to be used with prefetcher"""

    def __init__(self, url: str) -> None:
        """HTTP Client to be used with prefetcher

        :param url: url to send POST request to
        """
        self.url = url
        self.msg_recv = 0
        self.msg_sent = 0
        self.session = None

    async def send_message(self, request: Dict):
        """Sends a POST request to the server

        :param request: request as dict
        :return: send post message
        """
        return await self.session.post(url=self.url, json=request).__aenter__()

    async def __aenter__(self):
        """enter async context

        :return: start self
        """
        return await self.start()

    async def start(self):
        """Create ClientSession and enter context

        :return: self
        """
        import aiohttp

        self.session = aiohttp.ClientSession()
        await self.session.__aenter__()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.close(exc_type, exc_val, exc_tb)

    async def close(self, *args, **kwargs):
        """Close ClientSession

        :param args: positional args
        :param kwargs: keyword args"""
        await self.session.__aexit__(*args, **kwargs)


class HTTPBaseClient(BaseClient):
    """A MixIn for HTTP Client."""

    async def _get_results(
        self,
        inputs: InputType,
        on_done: Callable,
        on_error: Optional[Callable] = None,
        on_always: Optional[Callable] = None,
        **kwargs,
    ):
        """
        :param inputs: the callable
        :param on_done: the callback for on_done
        :param on_error: the callback for on_error
        :param on_always: the callback for on_always
        :param kwargs: kwargs for _get_task_name and _get_requests
        :yields: generator over results
        """
        with ImportExtensions(required=True):
            import aiohttp

        self.inputs = inputs
        request_iterator = self._get_requests(**kwargs)

        async with AsyncExitStack() as stack:
            try:
                cm1 = ProgressBar() if self.show_progress else nullcontext()
                proto = 'https' if self.args.https else 'http'
                url = f'{proto}://{self.args.host}:{self.args.port}/post'

                p_bar = stack.enter_context(cm1)
                iolet = await stack.enter_async_context(HTTPClientlet(url=url))

                prefetcher = HTTPClientPrefetcher(self.args, iolet=iolet)
                async for response in prefetcher.send(request_iterator):
                    r_status = response.status
                    r_str = await response.json()
                    if r_status == 404:
                        raise BadClient(f'no such endpoint {url}')
                    elif r_status < 200 or r_status > 300:
                        raise ValueError(r_str)

                    resp = Request(r_str)
                    resp = resp.as_typed_request(resp.request_type).as_response()
                    callback_exec(
                        response=resp,
                        on_error=on_error,
                        on_done=on_done,
                        on_always=on_always,
                        continue_on_error=self.continue_on_error,
                        logger=self.logger,
                    )
                    if self.show_progress:
                        p_bar.update()
                    yield resp
            except aiohttp.client_exceptions.ClientConnectorError:
                self.logger.warning(f'Client got disconnected from the HTTP server')



import numpy as np


class PCA:
    """:class:`PCA` is a class for dimensionality reduction using PCA in Jina.

    :param n_components: Number of components to keep when projecting with PCA
    :param whiten: Flag variable stating if there projecting with whitening
    """

    def __init__(self, n_components: int, whiten: bool = False):
        self.n_components = n_components
        self.whiten = whiten
        self.e_values = None
        self.w = None

    def fit(self, x_mat: np.ndarray):
        """
        Computes the projection matrix of the PCA algorithm and stores it to self.w

        :param x_mat: Matrix of shape (n_observations, n_features)
        """

        x_mat = x_mat - x_mat.mean(axis=0)

        # covariance matrix (n_features, n_features)
        cov = np.cov(x_mat.T) / x_mat.shape[0]

        # Compute eigenvalues eigenvectors of cov
        e_values, e_vectors = np.linalg.eig(cov)

        # Sort eigenvalues by magnitude (higher to lower)
        idx = e_values.argsort()[::-1]
        e_values = e_values[idx]

        # Sort eigen vectors by eigenvalue magnitude (higher to lower)
        e_vectors = e_vectors[:, idx]

        # Projection matrix contains eigenvectors sorted
        self.w = e_vectors
        self.e_values = e_values

    def transform(self, x_mat: np.ndarray) -> np.ndarray:
        """Projects data from n_features to self.n_components features.

        :param x_mat: Matrix of shape (n_observations, n_features)
        :return: Matrix of shape (n_observations, self.n_components)
        """
        if self.w is not None:
            x_mat_projected = x_mat.dot(self.w[:, : self.n_components])
            if self.whiten:
                return x_mat_projected / np.sqrt(self.e_values[0 : self.n_components])
            else:
                return x_mat_projected

    def fit_transform(self, x_mat: np.ndarray) -> np.ndarray:
        """Fits the PCA and returns a transformed data
        :param x_mat:  Matrix of shape (n_observations, n_features)
        :return: Matrix of shape (n_observations, self.n_components)
        """
        self.fit(x_mat)
        return self.transform(x_mat)



import datetime
import itertools
import math
import sys
import threading
import time
from collections import defaultdict
from functools import wraps
from typing import Optional, Union, Callable

from .. import __windows__
from jina.enums import ProgressBarStatus

from .logger import JinaLogger
from ..helper import colored, get_readable_size, get_readable_time


def used_memory(unit: int = 1024 * 1024 * 1024) -> float:
    """
    Get the memory usage of the current process.

    :param unit: Unit of the memory, default in Gigabytes.
    :return: Memory usage of the current process.
    """
    if __windows__:
        # TODO: windows doesn't include `resource` module
        return 0

    import resource

    return resource.getrusage(resource.RUSAGE_SELF).ru_maxrss / unit


def used_memory_readable() -> str:
    """
    Get the memory usage of the current process in a human-readable format.

    :return: Memory usage of the current process.
    """
    return get_readable_size(used_memory(1))


def profiling(func):
    """
    Create the Decorator to mark a function for profiling. The time and memory usage will be recorded and printed.

    Example:
    .. highlight:: python
    .. code-block:: python

        @profiling
        def foo():
            print(1)

    :param func: function to be profiled
    :return: arguments wrapper
    """
    from .predefined import default_logger

    @wraps(func)
    def arg_wrapper(*args, **kwargs):
        start_t = time.perf_counter()
        start_mem = used_memory(unit=1)
        r = func(*args, **kwargs)
        elapsed = time.perf_counter() - start_t
        end_mem = used_memory(unit=1)
        # level_prefix = ''.join('-' for v in inspect.stack() if v and v.index is not None and v.index >= 0)
        level_prefix = ''
        mem_status = f'memory Δ {get_readable_size(end_mem - start_mem)} {get_readable_size(start_mem)} -> {get_readable_size(end_mem)}'
        default_logger.info(
            f'{level_prefix} {func.__qualname__} time: {elapsed}s {mem_status}'
        )
        return r

    return arg_wrapper


class TimeDict:
    """Records of time information."""

    def __init__(self):
        self.accum_time = defaultdict(float)
        self.first_start_time = defaultdict(float)
        self.start_time = defaultdict(float)
        self.end_time = defaultdict(float)
        self._key_stack = []
        self._pending_reset = False

    def __enter__(self):
        _key = self._key_stack[-1]
        # store only the first enter time
        if _key not in self.first_start_time:
            self.first_start_time[_key] = time.perf_counter()
        self.start_time[_key] = time.perf_counter()
        return self

    def __exit__(self, typ, value, traceback):
        _key = self._key_stack.pop()
        self.end_time[_key] = time.perf_counter()
        self.accum_time[_key] += self.end_time[_key] - self.start_time[_key]
        if self._pending_reset:
            self.reset()

    def __call__(self, key: str, *args, **kwargs):
        """
        Add time counter.

        :param key: key name of the counter
        :param args: extra arguments
        :param kwargs: keyword arguments
        :return: self object
        """
        self._key_stack.append(key)
        return self

    def reset(self):
        """Clear the time information."""
        if self._key_stack:
            self._pending_reset = True
        else:
            self.accum_time.clear()
            self.start_time.clear()
            self.first_start_time.clear()
            self.end_time.clear()
            self._key_stack.clear()
            self._pending_reset = False

    def __str__(self):
        return ' '.join(f'{k}: {v:3.1f}s' for k, v in self.accum_time.items())


class TimeContext:
    """Timing a code snippet with a context manager."""

    time_attrs = ['years', 'months', 'days', 'hours', 'minutes', 'seconds']

    def __init__(self, task_name: str, logger: 'JinaLogger' = None):
        """
        Create the context manager to timing a code snippet.

        :param task_name: The context/message.
        :param logger: Use existing logger or use naive :func:`print`.

        Example:
        .. highlight:: python
        .. code-block:: python

            with TimeContext('loop'):
                do_busy()

        """
        self.task_name = task_name
        self._logger = logger
        self.duration = 0

    def __enter__(self):
        self.start = time.perf_counter()
        self._enter_msg()
        return self

    def _enter_msg(self):
        if self._logger:
            self._logger.info(self.task_name + '...')
        else:
            print(self.task_name, end=' ...\t', flush=True)

    def __exit__(self, typ, value, traceback):
        self.duration = self.now()

        self.readable_duration = get_readable_time(seconds=self.duration)

        self._exit_msg()

    def now(self) -> float:
        """
        Get the passed time from start to now.

        :return: passed time
        """
        return time.perf_counter() - self.start

    def _exit_msg(self):
        if self._logger:
            self._logger.info(
                f'{self.task_name} takes {self.readable_duration} ({self.duration:.2f}s)'
            )
        else:
            print(
                colored(
                    f'{self.task_name} takes {self.readable_duration} ({self.duration:.2f}s)'
                ),
                flush=True,
            )


class ProgressBar(TimeContext):
    """
    A simple progress bar.

    Example:
        .. highlight:: python
        .. code-block:: python

            with ProgressBar('loop'):
                do_busy()
    """

    col_width = 100
    clear_line = '\r{}\r'.format(' ' * col_width)
    spinner = itertools.cycle(['⠋', '⠙', '⠹', '⠸', '⠼', '⠴', '⠦', '⠧', '⠇', '⠏'])

    def __init__(
        self,
        description: str = 'Working...',
        message_on_done: Union[str, Callable[..., str], None] = None,
        final_line_feed: bool = True,
    ):
        """
        Create the ProgressBar.

        :param description: The name of the task, will be displayed in front of the bar.
        :param message_on_done: The final message to print when the progress is complete
        :param final_line_feed: if False, the line will not get a Line Feed and thus is easily overwritable.
        """
        super().__init__(description, None)
        self._bars_on_row = 40
        self._completed_progress = 0
        self._last_rendered_progress = 0
        self._num_update_called = 0
        self._on_done = message_on_done
        self._final_line_feed = final_line_feed
        self._stop_event = threading.Event()

    def update(
        self,
        progress: float = 1.0,
        description: Optional[str] = None,
        message: Optional[str] = None,
        status: ProgressBarStatus = ProgressBarStatus.WORKING,
        first_enter: bool = False,
    ) -> None:
        """
        Increment the progress bar by one unit.

        :param progress: The number of unit to increment.
        :param description: Change the description text before the progress bar on update.
        :param message: Change the message text followed after the progress bar on update.
        :param status: If set to a value, it will mark the task as complete, can be either "Done" or "Canceled"
        :param first_enter: if this method is called by `__enter__`
        """
        self._num_update_called += 0 if first_enter else 1
        self._completed_progress += progress
        self._last_rendered_progress = self._completed_progress
        elapsed = time.perf_counter() - self.start
        num_bars = self._completed_progress % self._bars_on_row
        num_bars = (
            self._bars_on_row
            if not num_bars and self._completed_progress
            else max(num_bars, 1)
        )
        num_fullbars = math.floor(num_bars)
        num_halfbars = 1 if (num_bars - num_fullbars <= 0.5) else 0

        if status in {ProgressBarStatus.DONE, ProgressBarStatus.CANCELED}:
            bar_color, unfinished_bar_color = 'yellow', 'yellow'
        elif status == ProgressBarStatus.ERROR:
            bar_color, unfinished_bar_color = 'red', 'red'
        else:
            bar_color, unfinished_bar_color = 'green', 'green'

        speed_str = (
            'estimating...'
            if first_enter
            else f'{self._num_update_called / elapsed:4.1f} step/s'
        )

        description_str = description or self.task_name or ''
        if status != ProgressBarStatus.WORKING:
            description_str = str(status)
        msg_str = message or ''
        self._bar_info = dict(
            bar_color=bar_color,
            description_str=description_str,
            msg_str=msg_str,
            num_fullbars=num_fullbars,
            num_halfbars=num_halfbars,
            speed_str=speed_str,
            unfinished_bar_color=unfinished_bar_color,
        )

    def _print_bar(self, bar_info):
        time_str = str(
            datetime.timedelta(seconds=time.perf_counter() - self.start)
        ).split('.')[0]
        sys.stdout.write(self.clear_line)
        sys.stdout.write(
            '{} {:>10} {:<}{:<} {} {} {}'.format(
                colored(next(self.spinner), 'green'),
                bar_info['description_str'],
                colored('━' * bar_info['num_fullbars'], bar_info['bar_color'])
                + (
                    colored(
                        '╸',
                        bar_info['bar_color']
                        if bar_info['num_halfbars']
                        else bar_info['unfinished_bar_color'],
                    )
                ),
                colored(
                    '━' * (self._bars_on_row - bar_info['num_fullbars']),
                    bar_info['unfinished_bar_color'],
                    attrs=['dark'],
                ),
                colored(time_str, 'cyan'),
                bar_info['speed_str'],
                bar_info['msg_str'],
            )
        )
        sys.stdout.flush()

    def _update_thread(self):
        sys.stdout.flush()
        while not self._stop_event.is_set():
            self._print_bar(self._bar_info)
            time.sleep(0.1)

    def _enter_msg(self):
        self.update(first_enter=True)

        self._progress_thread = threading.Thread(
            target=self._update_thread, daemon=True
        )
        self._progress_thread.start()

    def __exit__(self, exc_type, value, traceback):
        self.duration = self.now()

        self.readable_duration = get_readable_time(seconds=self.duration)

        if exc_type in {KeyboardInterrupt, SystemExit}:
            self._stop_event.set()
            self.update(0, status=ProgressBarStatus.CANCELED)
            self._print_bar(self._bar_info)
            return True  # prevent it from being propagated
        elif exc_type and issubclass(exc_type, Exception):
            self._stop_event.set()
            self.update(0, status=ProgressBarStatus.ERROR)
            self._print_bar(self._bar_info)
        else:
            # normal ending, i.e. task is complete
            self._stop_event.set()
            self._progress_thread.join()
            self.update(0, status=ProgressBarStatus.DONE)
            self._print_bar(self._bar_info)
            self._print_final_msg()

    def _print_final_msg(self):
        if self._last_rendered_progress > 1:
            final_msg = f'\033[K{self._completed_progress:.0f} steps done in {self.readable_duration}'
            if self._on_done:
                if isinstance(self._on_done, str):
                    final_msg = self._on_done
                elif callable(self._on_done):
                    final_msg = self._on_done()
            sys.stdout.write(final_msg)
            if self._final_line_feed:
                sys.stdout.write('\n')
        else:
            # no actual render happens
            sys.stdout.write(self.clear_line)
        sys.stdout.flush()
